# -*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

"""
====
NOTE
====

Please include all these imports in any new .py files you create, as they are necessary for the Sortal GI functions
"""

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import time

# Import for transformations
#from mpmath import pi, sin, cos
from copy import deepcopy

# Sortal GI methods import
import sortalgi as sgi

# Rhino Methods import
from sortalgi import rhino_methods as rm

# Sortal Library import
from sortal import sdlParser, sort, rule, form, flow

#id = rs.GetObject()
sort.cleanup()
rule.cleanup()
flow.cleanup()

"""
==============
DEMO STRUCTURE
==============
This demo demonstrates how to read from an SDL file to set up the sortal library, how to draw rules and shapes in the Rhino workspace based on those already
existing in the sortal library registers, how to create a flow sequence and apply it onto a shape, and how to save the flow sequence in an SDL file.

1. Reading SDL Files: Reading an SDL file and setting up sorts in the sort register
2. Drawing:
    a. Rules
    b. Shapes
3. Flow Creation: Creating a flow from rules
4. Apply a Flow: Applying a flow onto a shape
5. Saving Flows and Shapes: Save a flow/shape to a pre-existing SDL file

The grammar used in this demonstration is based on figures 20 and 21 in the following paper:
I. Jowers and C. Earl, 2011, Implementation of curved shape grammars, Environment and Planning B: Planning and Design 38, 616-635.
"""

"""
=================
READING SDL FILES
=================
Reads an SDL file and sets up the sort types inside the SDL file in the sort register
This also uploads any rules, shapes and flows (sequences of rules with additional looping and end conditions) to the sortal library registers for easy retrieval.
"""
print ('READING AN SDL FILE')
print ('Sort register before opening SDL file:')
print (sort.register)
print ()
print ('Rule register before opening SDL file:')
print (rule.register)
print ()

sgi.read_sdl('celtic_knots.sdl')

print ('Sort register after opening SDL file:')
print (sort.register)
print ()
print ('Rule register after opening SDL file:')
print (rule.register)
print ()
time.sleep(3)

"""
========================
DRAWING RULES AND SHAPES
========================
Draws rules and shapes in the Rhino workspace. Rules are drawn side by side and spaced apart from one another.
"""
# Draws a series of rules that will be relevant to the creation of flows later on
print ('VISUALIZING RULES AND SHAPES')
print ('The next few steps will help visualize some of the rules used for drawing Celtic knots. The rules are drawn side by side, apart from one another')
print ('If no reference point is selected, then the drawings will be generated and placed at the origin (0,0,0)\n')
refPoint = rs.GetObject('Select reference point for drawing rule 1 (labelled accordingly in "celtic_knots.3dm")')
sgi.draw_rule('r1', refPoint)

refPoint = rs.GetObject('Select reference point for drawing rule 2 (labelled accordingly in "celtic_knots.3dm")')
sgi.draw_rule('r2', refPoint)

refPoint = rs.GetObject('Select reference point for drawing rule 5 (labelled accordingly in "celtic_knots.3dm")')
sgi.draw_rule('r5', refPoint)

refPoint = rs.GetObject('Select reference point for drawing rule 9i (labelled accordingly in "celtic_knots.3dm")')
sgi.draw_rule('r9i', refPoint)

# Draws a shape 
refPoint = rs.GetObject('\nSelect reference point as origin point for shape (labelled accordingly in "celtic_knots.3dm")')
sgi.draw_shape('t0', refPoint)

"""
==============
FLOW CREATION
==============
Creates a flow (a sequence of rules with additional looping and end conditions) from existing rules in the sortal rule register
Stores the rule in the flow register

Notation:
    ( ) : Apply this sequence of rules first before moving on to the next step
    {<x>} : Apply this sequence of rules x number of times, x being an integer
    * : Apply rule onto shape as many times as possible (until a match cannot be found); this includes matching from 0 to n number of times
    
Note: 
    When naming a flow, the first character must be an alphabet character
"""
print ('\nCREATING A FLOW')
print ('The next step will show you how to create a flow sequence based on currently available rules in the rule register')
print ('Flow register after creating the flow sequence "my_1st_flow"')
print (flow.register)
print ()

flow_name = 'my_1st_flow'
flow_description = 'flow test'
flow_sequence = '(r1 r2*){2} r5 r9i*'
print ('Flow name: ' + flow_name)
print ('Flow description: ' + flow_description)
print ('Flow seqence: ' + flow_sequence)
sgi.create_flow(flow_name, flow_description, flow_sequence)

print ('Flow register after creating the flow sequence "my_1st_flow"')
print (flow.register)
print ()

"""
============
APPLY A FLOW
============
Applies the flow we just created onto a shape from the SDL file we opened at the start

The following notation is used in depicting the flow application onto a shape:
    (<integer A>/<integer B>) : integer A - index of rule application; integer B - number of total rule applications from applying rule onto shape
    ##<rule name>## >>backtrace<< : no applications found for rule and shape; backtrace indicates that one step is taken backwards to try other rule applications in the
    previous step before proceeding to the step where the sequence failed to find a match
"""
print ('\nAPPLYING A FLOW')
print ('The flow called "' + flow_name + '" will now be applied onto the initial shape "t0"')
print ('The steps of applying the rules onto the initial shape will be printed out one by one')
print ('A flow sequence applies each rule in a step on the result of the previous rule application')
refPoint = rs.GetObject('Select reference point for final result of flow application onto initial shape\n')
flow_result = sgi.apply_flow(flow_name, 't0', refPoint, hide = False)
sgi.create_shape('my_1st_flow_result', flow_result)

"""
============
SAVING FLOWS
============
Saves a flow object to the same SDL file we opened at the start
The flow called 'my_1st_flow' and the result of applying 'my_1st_flow' onto the shape 't0' are now stored inside 'celtic_knots.sdl'
"""
print ('\nSAVING A FLOW (AND A SHAPE)')
print ('The flow called "' + flow_name + '" and the shape called "my_1st_flow_result" will now be saved inside "celtic_knots.sdl"')
sgi.overwrite_sdl('celtic_knots.sdl', shapes = ['my_1st_flow_result'],  flows = [flow_name], append = True)

"""
=========
END NOTES
=========

There is another shape called 't1' inside 'celtic_knots.sdl' that is a long series of loops. You may experiment with applying different rules and flow sequences
onto this shape, as per the use of functions like sgi.find_rule_appns (demonstrated in the demo 'rule creation and application').
"""