#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r

# Sortal library imports
from sortal import sort

# API support code imports
from .convert.convert_sort_and_shape import convertSort

"""
=============
FUNCTION INFO
=============
'convert_sort' determines and constructs the counterpart sort type contained in sortalStructure, according to the target sort's 
specified parametric-dimension combination (non-parametric 2D - N2D, non-parametric 3D - N3D, parametric 2D - P2D, parametric 3D - P3D) 

INPUT
sortalStructure: Name (text string) or actual data structure of sort type to be converted over to new parametric-dimension structure;
                 this can be a disjunctive, attribute or primitive sort
targetSort: text string of target sort type; this may be 2D or 3D, non-parametric or parametric

For example:
    a. sortalStructure: meta3D, targetSort: 'P2D'; this will determine and construct the sort type that mirrors the geometry and their attributes in meta3D, 
    but as parametric 2D geometric sorts
    b. sortalStructure: lineSegP3D, targetSort: 'N2D'; this will construct the 2D line segment geometric sort and bring over any attributes present in lineSegP3D
    c. sortalStructure: ellipticalArc3D, targetSort: 'P3D'; this is not possible, as elliptical arcs are not enabled for parametric behavior in the sortal library

OUTPUT
targetSortType: New sort type corresponding to targetSort variation, that mirrors base geometry in sortalStructure, 
                but converts them to their counterparts as in targetSort, if successful
                Note: If the sortal structure being converted is a disjunctive sort, it is automatically included in 'rhino_shapes_c'; otherwise, it is only
                stored in the sort register
None, if unsuccessful

WARNINGS / ERRORS
TypeError: Input for sortalStructure is neither a text string nor a sort type data structure, 
           OR input for targetSort is neither a text string nor a sort type data structure
KeyError: Input for sortalStructure is a text string that is not present in the sort register
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: sortalStructure (sort type or name of sort type to be converted), targetSort (text string) 
           -> new sort type based on structure of sortalStruct and combination (2D, 3D, P2D, P3D) of targetSort
"""

def convert_sort(sortalStructure = None, targetSort = None, newSortName = 'newSort'):
    if 'rhino_shapes_c' in sort.register:
        targetSortType = None

        if sortalStructure is not None and targetSort is not None:
            if type(targetSort) != type('a'):
                raise TypeError('Input for target sort type data structure is neither a text string nor a sort type data structure')
                
            if type(sortalStructure) == type('a'):
                if sortalStructure in sort.register:
                    sortalStructure = sort.register[sortalStructure] 
                else:
                    raise KeyError('Sort type with the name ' + sortalStructure + ' is not present in the sort register.')
            else:
                try:
                    test = sortalStructure.type
                except:
                    raise TypeError('Input for sortal structure to be converted is neither the name of a sort type present in the sort register nor a sortal data type.')
                    
            targetSortType = convertSort(sortalStructure, targetSort, newSortName).sortType()
            return targetSortType
       
        else:
            if type(sortalStructure) != type('a'):
                try:
                    test = sortalStructure.type
                except:
                    raise TypeError('Input for sortal structure to be converted is neither the name of a sort type present in the sort register nor a sortal data type.')
            
            if type(targetSort) != type('a'):
                raise TypeError('Input for target sort dimension-parametric combination is not a text string. Only the following text string inputs are allowed: 2D, 3D, N2D, N3D, P2D, P3D.')
            return None

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                
