#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal Library Imports
from sortal import (rule, form, flow, sort, sdlParser, disjunctiveSort, compoundSort, primitiveSort, description)
            
# Python imports
from copy import deepcopy

"""
=============
FUNCTION INFO
=============
'save_to_sdl' stores rules, flows and an initial shape in an sdl file based on the variable fileName, by accepting their name references,
and retrieving the relevant sortal objects. If an sdl file with the same name already exists, 
then the user is prompted if they want to overwrite the file or give a new value to fileName.
The input for 'fileName' can also include the address of the intended location.

INPUT
fileName: Text string of file name (for .sdl file; this may include the intended location's full address)
rules: List of sortal rule object names (optional, this may be an empty list)
shape: One shape object name (optional, this may be ignored)
flows: List of sortal flow object names (optional, this may be an empty list)

Note: If including the full address of the target SDL file, the following format must be observed:
     'C:\\Users\AKIDRIBM\\AppData\\Roaming\\Grasshopper\\Libraries\\source_code_gh_dev\\'+ <sdl file name> +'.sdl'
     where every slash is doubled.

OUTPUT
If successful, an sdl file of name 'fileName' in the intended location with rule(s) and shape stored in it is created, and True is returned. Otherwise -
False: Unsuccessful operation

WARNINGS/ERRORS
MessageBox: If an SDL file with the same file name already exists in the location; the user is prompted if they would like to overwrite the file
            or give the current save_sdl function a new file name input
Warning: If the rule or shape or flow name does not exist in its corresponding register
TypeError: If the input for rules or flows is not a list or the input for shape is not a string; if there are any non-text inputs within
           the input for rules or flows
KeyError: If the shape name does not exist in the form register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up

Data Flow: file name, list of rule names, one shape, list of shape names 
           -> checking pre-existence of SDL file with the same name 
           -> if none, retrieval of sortal objects corresponding to rule names, shape name, flow names 
           -> SDL file creation in intended location
"""

def save_sdl(fileName = 'new.sdl', rules = [], shape = '', flows = []): 
    if 'rhino_shapes_c' in sort.register:
        go_ahead = False
        
        if type(rules) == type([]) and (type(shape) == type('a') or type(shape) == None):
            if len(rules) != 0 or len(shape) != 0 or len(flows) != 0:
                # Checks if file exists
                if os.path.exists(fileName):
                    go_ahead = rs.MessageBox("An sdl file with the same name in the same location already exists. Overwrite file?",  3, "File Overwrite")
        
                    if go_ahead == 6:
                        if '.sdl' not in fileName:
                            fileName = fileName + '.sdl'
                        
                    elif go_ahead == 7:
                        newName = raw_input('Please input new file name/directory path')
                        if '.sdl' not in newName:
                            fileName = newName + '.sdl'
                        
                    elif go_ahead == 2:
                        warnings.warn('Exiting save_sdl function')
                        return False
        
                else:
                    if '.sdl' not in fileName:
                        fileName = fileName + '.sdl'
                
                ruleObjects = []
                shapeObject = []
                flowObjects = []
                
                for ruleName in rules:
                    if type(ruleName) == type('a'):
                        if ruleName in rule.register:
                            ruleObjects.append(rule.register[ruleName])
                        else:
                            warnings.warn(ruleName + ' does not exist in rule register.')
                    else:
                        try:
                            if ruleName.type == 'rule':
                                ruleObjects.append(ruleName)
                        except:
                            raise TypeError(('Invalid rule input', ruleName))
                
                if type(shape) == type('a'):
                    if shape in form.formRegister:
                        shapeObject.append(form.formRegister[shape])
                    else:
                        if shape == '':
                            pass
                                
                        else:
                            warnings.warn(shape + ' does not exist in form register. No shape will be included in the resulting SDL file.')
                else:
                    try: 
                        if 'form' in shape.type:
                            shapeObject.append(shape)
                    except:
                        raise TypeError(('Invalid shape input', shape))
                 
                for flowName in flows:
                    if type(flowName) == type('a'):
                        if flowName in flow.register:
                            flowObjects.append(flow.register[flowName])
                        else:
                            warnings.warn(flowName + ' does not exist in the flow register.')
                    else:
                        try:
                            if flowName.type == 'flow':
                                flowObjects.append(flowName)
                        except:
                            raise TypeError(('Invalid flow input', flowName))
    
                sdlParser.writeGrammar(fileName, ruleObjects, shapeObject, flowObjects)
                    
                # file has been created
                if shape != '':
                    message = (fileName + ' has been created with ' + str(len(ruleObjects)) + ' rule/s, ' + str(len(flowObjects)) + ' flow/s and the shape ' + shape)
                else:
                    message = (fileName + ' has been created with ' + str(len(ruleObjects)) + ' rule/s and ' + str(len(flowObjects)) + ' flow/s')
                    
                print (message)
                return True
                
            else:
                warnings.warn('All inputs (rules, shape, flows) are empty. No SDL file will be created.')
                return False
            
        else:
            if type(rules) != type([]):
                raise TypeError('Rules input is not a list')
                
            if type(flows) != type([]):
                raise TypeError('Flows input is not a list')
                
            if type(shape) != type('a'):
                raise TypeError('Shape input is not a string')
                
            elif shape not in form.formRegister:
                raise KeyError('Shape input does not exist in form register')

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
           
           
            