#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# API support code imports
from .convert.convert_agnostic_to_sortal import convertA
from .convert.convert_sortal_to_agnostic import convertS
from ..rhino_specific.visualAgnostic import visualA
from ..rhino_specific.rhino_specific_methods import rhino_methods as rm

# Sortal library imports
from sortal import rule, form, sort

# Conversion code initializations
convertA = convertA()
convertS = convertS()

"""
=============
FUNCTION INFO
=============
'find_rule_appns' finds rule applications from a subshape and shape, or from a set of GUIDs.

SUPPORT
construct_descriptions: Constructs description string as text string ready for re-input into the creation of a new shape using sgi.create_shape

INPUT
Required
chosenRule : Rule name as text string
shape: Name of main shape as text string as recorded in sortal library's form register pertaining to shape 

Optional
subshape: Name of subshape as text string as recorded in sortal library's form register pertaining to subshape 
refPt: GUID of point, tuple, or Rhino Geometry point of reference point which will serve as the 'origin' with which shapes 
       will be plotted in respect to (in principle, a vector)
layerName: Name of target layer as text string within Rhino workspace where the GUIDs will be drawn onto
prnt: Boolean value True/False; True - prints out description individuals as input-ready text string (to create_shape)
      as well as printE form of shape after rule application; False (default) - nothing is printed
shapeIds: List of Rhino GUIDs to clear after new shapes have been drawn 

OUTPUT
appnGeometry: List of lists of Rhino GUIDs (each list correspond to a shape after a certain rule application), if successful;
              these geometries are hidden from the viewport, by default
None, if unsuccessful

WARNINGS / ERRORS
TypeError: If initial shape or subshape is not a list of GUIDs or a text string or is empty, or if the input for chosenRule is not a text string
KeyError: If initial shape or subshape input's name is not present in the form register, or if rule name is not present in the rule register
          or if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
Warning: If subshape is not part of shape -> exits function and returns False

Data Flow:  Names of rule, shape
            -> detect rule applications within shape
            -> obtain sortal rule applications and convert the resulting shapes to agnostic form
            (-> rule applications based on subshape, subshapeAgnostic if subshape is specified)
            -> draw each rule application and construct its external descriptions text string
            -> hide Rhino geometries of rule applications, print external description text string if prnt set to True
"""

def find_rule_appns(chosenRule, shape, subshape = None, refPt = r.Geometry.Point3d(0,0,0), layerName = 'Default', hide = True, prnt = False, shapeIds = None):
    if 'rhino_shapes_c' in sort.register:
        # RETRIEVAL OF RULE OBJECT
        # Checks if rule input is a a valid string input that corresponds to a rule in the rule register
        if type(chosenRule) == type('a'):
            if chosenRule in rule.register:
                chosenRule = rule.register[chosenRule]
            else:
                raise KeyError('Rule name is not present in the rule register.')
        else:
            raise TypeError('Rule name input is not a text string.')
    
        ruleAppns = []          # List of rule applications to be returned (either as agnostic dict or Rhino GUIDs)
        subshapeSortal = None
        
        # Text string reference to shape
        # Retrieves shape of the corresponding name from the sortal form register
        if type(shape) == type('a'):
            if shape in form.formRegister:
                shapeSortal = form.formRegister[shape]
            else:
                raise KeyError(('Shape with the name ' + shape + ' does not exist in the sortal form register.'))
        
        else:
            raise TypeError('Shape name input is not a text string.')
            
        # RETRIEVAL OF SUBSHAPE
        # Checks if subshape is valid (optional)
        # Retrieves subshape from sortal form register if necessary
        if type(subshape) == type('a'):
            if type(subshape) == type('a'):
                if subshape in form.formRegister:
                    subshapeSortal = form.formRegister[subshape]
                else:
                    raise KeyError('Shape with the name', subshape, 'does not exist in the sortal form register.')
                
        else:
            if subshape != None:
                raise TypeError('Subshape name input is not a text string.')
        
        # COMPATIBILITY BETWEEN SUBSHAPE AND SHAPE
        # If subshape is provided
        if subshapeSortal != None:
            
            # Checks if subshape is part of main shape before proceeding
            if subshapeSortal.partOf(shapeSortal):
                # RULE APPLICATION DETECTION
                subshapeApps = chosenRule.detect(subshapeSortal)       # Generates matches, results and rule applications for subshape
                
                # CONVERSION TO AGNOSTIC DICTIONARY
                for index1 in range(len(subshapeApps)):
                    after = convertS.toAgnostic(subshapeApps[index1].perform(shapeSortal.duplicate()))  # Entire shape after rule application
                    ruleAppns.append(after)
                        
            else:   
                warnings.warn('Subshape is not part of main shape. Matches cannot be found. Exiting function.')
                return None
                
        # Subshape is not provided
        else:
            # RULE APPLICATION DETECTION
            shapeApps = chosenRule.detect(shapeSortal)       # Generates matches, results and rule applications for subshape
            
            # CONVERSION TO AGNOSTIC DICTIONARY
            for index1 in range(len(shapeApps)):
                after = convertS.toAgnostic(shapeApps[index1].perform(shapeSortal.duplicate())) # Entire shape after rule application
                ruleAppns.append(after)
                
        # Converts agnostic dictionaries inside ruleAppns to a list of Rhino GUIDs (each shape corresponds to one list)
        appnGeometry = []
        
        # PLOTTING OF RULE APPLICATIONS
        for row_count, ruleAppn in enumerate(ruleAppns):
            vis = visualA(ruleAppn, refPt, layerName)
            newShapeIDs = vis.plotAgnostic()                       # Redraws shape after rule application based on agnostic dictionary input 'ruleAppn'  
            appnGeometry.append(newShapeIDs)
            descriptions = None
            
            if 'descriptions' in ruleAppn:
                description_text = ruleAppn['descriptions']
                descriptions = construct_description(description_text)
            
            if prnt:
                # Display text
                print ('Rule application ' + str(row_count+1) + ':')
                if descriptions != None:
                    print ('\nApplication description(s) text string')
                    print (descriptions + ' \n')
    
                convertA.toSortal(ruleAppn).printE()
                print('\n')
            
            if hide:
                # Hides geometry from Rhino viewport
                rs.HideObjects(newShapeIDs)
            
        # CLEARING SHAPE
        # Clears previous shape IDs inside frame; input is a list of GUIDs
        if type(shapeIds) != type(None):
            rm.clearShape(shapeIds)  

        return appnGeometry

    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')

"""
================
SUPPORT FUNCTION
================
'construct_description' constructs the string text for any external, standalone descriptions present in a rule application

INPUT
descriptions: Dictionary containing description individuals grouped according to the name of their description sort type

OUTPUT
description_text: Text string assembling description individuals based on their description sort type and data; this text string
                  may be copy-pasted as direct input for future shape creation
"""

def construct_description(descriptions):
    description_text = ''

    for desc_name in descriptions:
        desc_text = descriptions[desc_name]
        
        if description_text == '':
            description_text = desc_name + ':'
        else:
            description_text = description_text+ '|' + desc_name + ':'
        
        for count, desc in enumerate(desc_text):
            if count == 0:
                description_text = description_text + desc
            else:
                description_text = description_text +';' + desc
                
    return description_text