#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

import rhinoscriptsyntax as rs
from builtins import range
import scriptcontext as sc
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, lineSegment2D, 
                    label, rWeight, disjunctiveSort, attributeSort, affineTransform,  vector3D, sort, key)

from mpmath import pi, sin, cos

"""
ConvertS CLASS
Takes sortal metaform as input and converts it into an agnostic form

INPUT
Any metaform, discrete form or interval form; these are sortal objects

RETURNS
Agnostic dictionary form of sortal object with format: {'points': [<p1>, <p2>], 'line segments': [<ls1>, <ls2>]}
""" 

class convertS:
    
    def __init__(self):
        pass

    # Converts from sortal compound form to compound agnostic object
    def toAgnostic(self, form):
        self.form = form
        self.type = form.type
        self.compoundAO = {}
        self.objects = {}
        self.classification = ''
        sc.sticky['mainSort'] = self.form.sort.name
        
        # Entered if form is a metaform, i.e. contains both points and line segments
        # Structure of metaform: <metaform name>.elements contains one form of line segments, one form of points
        # To access each individual point or line segment -> since the list stored in the metaform 
        # is constructed as [<point form>, <line segment form>]
        # i.e. <metaform>.name.elements[0].elements -> returns list of points
        # <metaform>.name.elements[1].elements -> returns list of line segments
        # Resutlting agnostic object is stored in self.objects
        
        # COMPOUND SORT
        if self.type == 'form-metaForm': 
            formElements = self.form.elements  
            check = False
            
            # Determines if it is necessary to place the resulting agnostic dictionary in a compound object agnostic dictionary
            # If one of the sorts is a dummy sort made of two description sort types, then the agnostic dictionary output reflects a disjunctive sort structure
            if len(sort.register['rhino_shapes_c'].component) == 2:
                for sortType in sort.register['rhino_shapes_c'].component:
                    descCount = 0
                    if sortType.type == 'sort-disjunctive':
                        for component in sortType.component:
                            if 'description' in component.baseSort().characteristic.type:
                                descCount += 1

                        if descCount == len(sortType.component):
                            check = True

                    if check:
                        break
                    
            # Handles disjunctive sorts inside compound agnostic object
            if formElements[0].type == 'form-metaForm':
                for formElement in formElements:
                    name = formElement.sort.name.split('_')
                    name = name[-1]
                    self.objects['descriptions'] = {}
                    self.fMetaform(formElement, formElement.sort.name)

                    # Creates new entry in compound agnostic object
                    if name not in self.compoundAO:
                        self.compoundAO[name] = []

                    if check == False:
                        self.compoundAO[name] = self.objects
                
                if len(self.objects['descriptions']) == 0:
                    del self.objects['descriptions'] 
                    
                if check:
                    self.finalObject = self.objects
                else:
                    self.finalObject = self.compoundAO

            # Handles single disjunctive sort
            else:
                self.fMetaform(formElements, self.form.sort.name)
                name = self.classification
                self.finalObject = self.objects
        
        else:
            # SIMPLE SORTS 
            # Entered if form only contains points/circles/ellipses 
            if self.type == 'form-discrete':
                self.fDiscrete(self.form.elements, self.form.sort.name)
               
            # Entered if form only contains line segments
            # Extracts data from line segments; parses them into agnostic object data
            elif self.type == 'form-interval':
               self.fInterval(self.form.elements, self.form.sort.name)

            # Entered if form only contains plane segments 
            # Extracts data from plane segments; parses them into agnostic object data
            elif elements[ind].type == 'form-coForm' or elements[ind].type == 'form-areal':
                self.fAreal(individuals, self.form.sort.name)
            
            # Circular/elliptical arcs
            elif elements[ind].type == 'form-cyclical':
                self.fCyclical(individuals, self.form.sort.name)
            
            # Bezier Curves
            elif elements[ind].type == 'form-curve':
                self.fCurve(individuals, self.form.sort.name)
                
            self.finalObject = self.objects

        return self.finalObject
    
    def fMetaform(self, elements, sortName):
        for a in sort.register[sortName].component:
            if a.name not in self.objects and 'description' not in sort.register[a.name].baseSort().characteristic.type:
                self.objects[a.name] = []
                
        for individual in elements:
            if len(individual.elements) > 0:
                if individual.type == 'form-metaForm':
                    self.fMetaform(individual.elements, individual.sort.name)
                    
                # Converts points/circles/points
                elif individual.type == 'form-discrete':
                    self.fDiscrete(individual.elements, individual.sort.name)
                            
                # Converts lines
                elif individual.type == 'form-interval':
                    self.fInterval(individual.elements, individual.sort.name)   # Extracts data from line segments; parses them into agnostic object data
                
                # Converts circular/elliptical arcs
                elif individual.type == 'form-cyclical':
                    self.fCyclical(individual.elements, individual.sort.name)
                
                # Converts bezier curves
                elif individual.type == 'form-curve':
                    self.fCurve(individual.elements, individual.sort.name)
                    
                # Converts plane segments
                elif individual.type == 'form-coForm' or individual.type == 'form-areal':
                    self.fAreal(individual.elements, individual.sort.name)     # Extracts data from plane segments; parses them into agnostic object data
        
        return
        
    def fInterval(self, individuals, sortName):
        if len(individuals) > 0:
            self.objects[sortName] = self.Line(individuals)
    
            if self.classification == '':
                name = individuals[0].sort.baseSort().characteristic.type
                self.classification = name[-2:]
                
                if 'P' in name:
                    self.classification = 'P'+ self.classification
        else:
            self.objects[sortName] = []   
            
        return
    
    # Handles form-discrete: Descriptions and/or Points
    def fDiscrete(self, individuals, sortName):
        if len(individuals) > 0:
            for indiv in individuals:
                obj = []
                
                # For descriptions
                if indiv.type == 'ind-description':
                    # Creates new dictionary item with name of description sort type as key name
                    if indiv.sort.baseSort().name not in self.objects['descriptions']:
                        self.objects['descriptions'][indiv.sort.baseSort().name] = []
                    
                    # Appends description individual to corresponding dictionary item inside self.objects['descriptions']
                    self.objects['descriptions'][indiv.sort.baseSort().name].append(indiv.getValue()[1:-1])
                
                elif 'circle' in indiv.type:
                    if type (indiv) == type(self.form.elements):
                        for ind in indiv:
                            self.objects[sortName].append(self.Circle(ind))
                    else:
                         self.objects[sortName].append(self.Circle(indiv))
                    
                elif 'ellipse' in indiv.type:
                    if type (indiv) == type(self.form.elements):
                        for ind in indiv:
                            self.objects[sortName].append(self.Ellipse(ind))
                    else:
                         self.objects[sortName].append(self.Ellipse(indiv))
                
                # Extracts data from points
                else:
                    if type (indiv) == type(self.form.elements):
                        for ind in indiv:
                            self.objects[sortName].append(self.Point(ind))
                    else:
                         self.objects[sortName].append(self.Point(indiv))
            
            if len(individuals) > 0 and self.classification == '' and indiv.sort.baseSort().characteristic.type != 'ind-description':
                name = indiv.sort.baseSort().characteristic.type
                self.classification = name[-2:]
    
                if 'P' in name:
                    self.classification = 'P'+ self.classification
        else:
            self.objects['descriptions'] = []
            self.objects[sortName] = []
            
        return
        
    def fAreal(self, individuals, sortName):
        if len(individuals) > 0:
            for indiv in individuals:
                if '2D' in indiv.sort.name:
                    obj = self.Plane(indiv)
                    self.objects[sortName].append(obj)
                elif '3D' in indiv.sort.name:
                    for planeSegment in indiv.elements:
                        obj = self.Plane(planeSegment)
                        self.objects[sortName].append(obj)
                 
            if '2D' or '3D' in individuals[0].sort.baseSort().characteristic.type:
                name = individuals[0].sort.baseSort().characteristic.type[-2:]
                self.classification = name[-2:]
    
                if 'P' in name:
                    self.classification = 'P'+ self.classification
        else:
            self.objects[sortName] = []
            
        return
    
    def fCyclical(self, individuals, sortName):
        if len(individuals) > 0:
            for indiv in individuals:
                if 'circular' in sort.register[sortName].baseSort().characteristic.type:
                    self.objects[sortName].append(self.cArc(indiv))
                elif 'elliptical' in sort.register[sortName].baseSort().characteristic.type:
                    self.objects[sortName].append(self.eArc(indiv))
                
            if '2D' or '3D' in individuals[0].sort.baseSort().characteristic.type:
                name = individuals[0].sort.baseSort().characteristic.type[-2:]
                self.classification = name[-2:]
    
                if 'P' in name:
                    self.classification = 'P'+ self.classification
        else:
            self.objects[sortName] = []
            
        return
        
    def fCurve(self, individuals, sortName):
        if len(individuals) > 0:
            for indiv in individuals:
                obj = self.Bezier(indiv)
                self.objects[sortName].append(obj)
                
            if '2D' or '3D' in individuals[0].sort.baseSort().characteristic.type:
                name = individuals[0].sort.baseSort().characteristic.type[-2:]
                self.classification = name[-2:]
    
                if 'P' in name:
                    self.classification = 'P'+ self.classification
        else:
            self.objects[sortName] = []
            
        return
    
    """
    ==========================================
    EXTRACTING DATA FROM GEOMETRIC INDIVIDUALS
    ==========================================
    Functions for extracting geometric/attribute data from sortal individualss
    """
    
    # Extracts data from line segments, returns agnostic tuple ((coordinates), {attributes dictionary}) of line segment data
    def Line(self, line_set):
        v1, v2 = 0, 0
        lines = []
        
        # Extracts attributes and coordinates
        for i, indiv in enumerate(line_set):
            if '2D' in indiv.type:
                v1 = (float(indiv.start.x.value), float(indiv.start.y.value),)
                v2 = (float(indiv.end.x.value), float(indiv.end.y.value),)
            elif '3D' in indiv.type:
                v1 = (float(indiv.start.x.value), float(indiv.start.y.value), float(indiv.start.z.value),)
                v2 = (float(indiv.end.x.value), float(indiv.end.y.value), float(indiv.end.z.value),)
                
            coords = (v1, v2)
            att = {}
            
            # Extracts attributes from line segments inside line_set if there are attributes inside
            # Appends them to text and weight lists with index number corresponding to which line they belong to
            # form-discrete - labels, descriptions
            # form-ordinal - weights
            # form-special - special variant of weights specifically for Rhino, 
            # does operations arithmetically (color, weight)
            
            if line_set[i].hasAttribute():
                ind  =  line_set[i].attribute
                att = self.createAtt('ls', ind)
                
            if line_set[i].key != None:
                att['tags'] = line_set[i].key.name 
           
            lines.append((coords, att))

        return lines
    
     
    # Extracts data from points, returns agnostic tuple ((coordinates), {attributes dictionary}) of point data
    def Point(self, pointSource):
        if '2D' in pointSource.type:
            pt = (float(pointSource.value.x.value), float(pointSource.value.y.value),)
        elif '3D' in pointSource.type:
            pt = (float(pointSource.value.x.value), float(pointSource.value.y.value), float(pointSource.value.z.value))
            
        att = {}
        
        # Handles point if it has attributes
        if pointSource.hasAttribute():
            ind  =  pointSource.attribute # Assigns pointSource's attribute
            att = self.createAtt('pt', ind)
                            
        pt = ((pt), att)
            
        return pt
        
    # Extracts data from points, returns agnostic tuple ((coordinates), {attributes dictionary}) of point data
    def Plane(self, planeSource):
        coords = {}
        coords['outer'] = []
        coords['inner'] = []

        for count, bound in enumerate(planeSource.boundaries):
            if count == 0:
                for vector in bound.polygonal:
                    if '2D' in vector.type:
                        coords['outer'].append((vector.x, vector.y))
                    elif '3D' in vector.type:
                        coords['outer'].append((vector.x, vector.y, vector.z))
                    
            else: 
                innerBound = ()
                for vector in bound.polygonal:
                    if '2D' in vector.type:
                        innerBound = innerBound + ((vector.x, vector.y),)
                    elif '3D' in vector.type:
                        innerBound = innerBound + ((vector.x, vector.y, vector.z),)
                        
                coords['inner'].append(innerBound)
        
        # Handles point if it has attributes
        if planeSource.hasAttribute():
            ind  =  planeSource.attribute # Assigns pointSource's attribute
            att = self.createAtt('pl', ind)

        plane = (coords, att)
        
        return plane
        
    def Circle(self, circleSource):
        center = circleSource.center
        radius = circleSource.radius
        circle_points = circleSource.getPoints()

        if '2D' in circleSource.type:
            coords = ((float(center.x), float(center.y)), float(radius), (float(circle_points[0].x), float(circle_points[0].y)), (float(circle_points[1].x), float(circle_points[1].y)))
        else:
            coords = ((float(center.x), float(center.y), float(center.z)), float(radius), (float(circle_points[0].x), float(circle_points[0].y), float(circle_points[0].z)), (float(circle_points[1].x), float(circle_points[1].y), float(circle_points[1].z)))
            #coords = ((float(center.x), float(center.y), float(center.z)), float(radius), (float(start.x), float(start.y), float(start.z)), (float(end.x), float(end.y), float(end.z)))
        
        if circleSource.hasAttribute():
            ind = circleSource.attribute
            att = self.createAtt('c', ind)
        
        circle = (coords, att)
        
        return circle
        
    def Ellipse(self, ellipseSource):
        center = ellipseSource.center
        foc1 = ellipseSource.foci[0]
        foc2 = ellipseSource.foci[1]
        pin_pt = ellipseSource.getPoints()[2]
        
        if '2D' in ellipseSource.type:
            coords = ((center.x, center.y), (float(foc1.x), float(foc1.y)), (float(foc2.x), float(foc2.y)), (float(pin_pt.x), float(pin_pt.y)))
        else:
            coords = ((center.x, center.y, center.z), (float(foc1.x), float(foc1.y), float(foc1.z)), (float(foc2.x), float(foc2.y), float(foc2.z)), (float(pin_pt.x), float(pin_pt.y), float(pin_pt.z)))
        
        if ellipseSource.hasAttribute():
            ind = ellipseSource.attribute
            att = self.createAtt('e', ind)
        
        ellipse = (coords, att)
        
        return ellipse
        
    def cArc(self, cArcSource):
        center = cArcSource.center
        start = cArcSource.start
        end = cArcSource.end
        pin = cArcSource.pin
        
        if '2D' in cArcSource.type:
            coords = ((float(center.x), float(center.y)), (float(start.x), float(start.y)), (float(end.x), float(end.y)), (float(pin.x), float(pin.y)))
        else:
            coords = ((float(center.x), float(center.y), float(center.z)), (float(start.x), float(start.y), float(start.z)), (float(end.x), float(end.y), float(end.z)), (float(pin.x), float(pin.y), float(pin.z)))
            
        if cArcSource.hasAttribute():
            ind = cArcSource.attribute
            att = self.createAtt('cArc', ind)
            
        cArc = (coords, att)
        
        return cArc
        
    def eArc(self, eArcSource):
        foc1 = eArcSource.foci[0]
        foc2 = eArcSource.foci[1]
        start = eArcSource.start
        end = eArcSource.end
        pin = eArcSource.pin
        short_pt = e.getQuadPoints()[0]
        long_pt = e.getQuadPoints()[1]
        
        if '2D' in eArcSource.type:
            coords = ((float(foc1.x), float(foc1.y)), (float(foc2.x), float(foc2.y)), (float(start.x), float(start.y)), (float(end.x), float(end.y)), (float(pin.x), float(pin.y)), (short_pt.x, short_pt.y), (long_pt.x, long_pt.y))
        else:
            coords = ((float(foc1.x), float(foc1.y), float(foc1.z)), (float(foc2.x), float(foc2.y), float(foc2.z)), (float(start.x), float(start.y), float(start.z)), (float(end.x), float(end.y), float(end.z)), (float(pin.x), float(pin.y),  float(pin.z)), (short_pt.x, short_pt.y, short_pt.z), (long_pt.x, long_pt.y, long_pt.z))
            
        if eArcSource.hasAttribute():
            ind = eArcSource.attribute
            att = self.createAtt('eArc', ind)
            
        eArc = (coords, att)
        
        return eArc
        
    def Bezier(self, bezierSource):
        start = bezierSource.start
        end = bezierSource.end
        ctrl = bezierSource.controlPoint
        att = {}
        
        if '2D' in bezierSource.type:
            coords = ((float(start.x), float(start.y)), (float(ctrl.x), float(ctrl.y)), (float(end.x), float(end.y)))
        else:
            coords = ((float(start.x), float(start.y), float(start.z)), (float(ctrl.x), float(ctrl.y), float(ctrl.z)), (float(end.x), float(end.y), float(end.z)))
            
        if bezierSource.hasAttribute():
            ind = bezierSource.attribute
            att = self.createAtt('b', ind)
        
        bezier = (coords, att)
        
        return bezier
        
    def createAtt(self, geomType, ind):
        att = {}
        weightCheck = ['ls', 'c', 'cArc', 'e', 'eArc', 'b']
        
        # Extracts label or description
        if ind.type == 'form-discrete':
            if ind.elements[0].getValue()[0] == '`':
                att['label_descriptions'] = [ind.elements[0].getValue()[1:-1]]
                          
            else:
               att['labels'] = [ind.elements[0].getValue()]

        # Extracts color of point/text dot OR enumerative value OR weight255
        elif ind.type == 'form-ordinal' or ind.type == 'form-special':
            if type(ind.individual.getValue()) == type(1.0) or type(ind.individual.getValue()) == type(100):
                if geomType in weightCheck:
                    att['weight'] = ind.individual.getValue()
                else:
                    att['weight255'] = [ind.individual.getValue(), ind.individual.getValue(), ind.individual.getValue()]
                
            elif type(ind.individual.getValue()) == type([]):
                att['color'] = [ind.individual.getValue()[0], ind.individual.getValue()[1], ind.individual.getValue()[2]]

            elif type(ind.individual.getValue()) == type('a') or type(ind.individual.getValue()) == type(1):
                att['enums'] = ind.individual.getValue()[1:-1]

        # Handles object if it has several attributes
        elif ind.type == 'form-metaForm':
            for attribute in ind.elements:
                if attribute.type == 'form-discrete':
                    if attribute.elements[0].getValue()[0] == '`':
                        att['label_descriptions'] = [attribute.elements[0].getValue()[1:-1]]
                      
                    else:
                        att['labels'] = [attribute.elements[0].getValue()]
                           
                if attribute.type == 'form-ordinal' or attribute.type == 'form-special':
                    if type(attribute.individual.getValue()) == type(100) or type(attribute.individual.getValue()) == type(1.0):
                        if geomType in weightCheck:
                            att['weight'] = [attribute.individual.getValue(), attribute.individual.getValue(), attribute.individual.getValue()]
                        else:
                            att['weight255'] = [attribute.individual.getValue(), attribute.individual.getValue(), attribute.individual.getValue()]

                    elif type(attribute.individual.getValue()) == type([]):
                        att['color'] = [attribute.individual.getValue()[0], attribute.individual.getValue()[1], attribute.individual.getValue()[2]]

                    elif type(attribute.individual.getValue()) == type('a'):
                        att['label_descriptions'] = [attribute.individual.getValue()]
                    
        return att
        
