"""

# -*- coding: utf-8 -*-

@author: akidribm
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

                    
"""
=====================
MANUAL SETUP of SORTS
=====================
This example demonstrates how to set up primitive, attribute, disjunctive sorts (and a compound sort as well)
"""

from sortal import (primitiveSort, point2D, point3D, pointP2D, pointP3D, 
                    lineSegment2D, lineSegment3D, lineSegmentP2D, lineSegmentP3D, 
                    planeSegment2D, planeSegment3D, planeSegmentP2D, planeSegmentP3D,
                    color, label, rWeight, enumerative, disjunctiveSort, 
                    attributeSort,  description, sort, compoundSort)
                    
"""
=========
FILE INFO
=========

Defines primitive sort types
These are used to compose the discursive sort that feeds into parse_disjunctive_sort 
and other applications, e.g. create_rule.py, apply_rule.py
"""

"""
===============
PRIMITIVE SORTS
===============
Creates primitive geometric sorts in all available variants: nonparametric 2D/3D, parametric 2D/3D
Sort names are not fixed and can be adapted to user's naming conventions
"""
# 2D SORTS
# Parametric geometric sorts
pointSortP2D = primitiveSort('pointP2D', pointP2D)
lineSegSortP2D = primitiveSort('lineSegP2D', lineSegmentP2D)
planeSegSortP2D = primitiveSort('planeSegP2D', planeSegmentP2D) 

# Non-parametric geometric sorts
pointSort2D = primitiveSort('point2D', point2D)
lineSegSort2D = primitiveSort('lineSeg2D', lineSegment2D)
planeSegSort2D = primitiveSort('planeSeg2D', planeSegment2D) 

# 3D SORTS
# Parametric geometric sorts
pointSortP3D = primitiveSort('pointP3D', pointP3D)
lineSegSortP3D = primitiveSort('lineSegP3D', lineSegmentP3D)
planeSegSortP3D = primitiveSort('planeSegP3D', planeSegmentP3D) 

# Non-parametric geometric sorts
pointSort3D = primitiveSort('point3D', point3D)
lineSegSort3D = primitiveSort('lineSeg3D', lineSegment3D)
planeSegSort3D = primitiveSort('planeSeg3D', planeSegment3D) 

"""
=========================
PRIMITIVE ATTRIBUTE SORTS
=========================
Creates primitive attribute sorts such as labels, weights (of different behaviors and min/max values), descriptions, enumeratives
Sort names are not fixed and can be adapted according to user's naming conventions
"""

arg = '{black,white}[black white white white]'
enumSort = primitiveSort('enumerative', enumerative, arg)

# Weight attribute definitions
# rWeight or weight can be used, with the crucial difference being
# rWeight - arithmetic operations on values
# weight - ordinal operations on values
# rWeight does not require an upper limit, but one may be specified
# This upper limit weight value is the value all weightSort individuals are maximised to 
# when any value above the former is stored in the sortal library
weightSort = primitiveSort('weight', rWeight)
weight255Sort = primitiveSort('weight255', rWeight, '255')

# Other attribute types
labelSort = primitiveSort('label', label)
descSort = primitiveSort('description', description)
ptDSort = primitiveSort('ptD', description)
lsDSort = primitiveSort('lsD', description)
plDSort = primitiveSort('plD', description)
colorSort = primitiveSort('color',color,'RGB RGB_SUM,')

"""
========================
CREATING ATTRIBUTE SORTS
========================
To create an attribute sort (baseSort ^ attributes), the disjunctive attributes sort (containing all attributes to be attached to a geometric base sort)
must first be created
"""

# 1. Create the disjunctive sort containing all pertinent attributes
# Name of disjunctive sort with pertinent attributes
attrib_sort_1 = 'label_lsD_color'
attrib_sort_2 = 'desc_plD_weight255'

# List containing pertinent attributes
attrib_sorts_1 = [labelSort, lsDSort, colorSort]
attrib_sorts_2 = [descSort, plDSort, weight255Sort]

# Creation of disjunctive sort: disjunctiveSort (<sort_name>, [list of sorts])
attribSort_1 = disjunctiveSort(attrib_sort_1, attrib_sorts_1)
attribSort_2 = disjunctiveSort(attrib_sort_2, attrib_sorts_2)

# 2. Create the attribute sort, with a geometric base sort and with the disjunctive attributes sort
# attributeSort(<sort_name>, base_geometric_sort, disj_attributes_sort)
attribute_sort_1_d = attributeSort(lineSegSort3D.name + '_' +  attrib_sort_1, lineSegSort3D, attribSort_1)
attribute_sort_2_d = attributeSort(lineSegSort3D.name + '_' +  attrib_sort_2, lineSegSort3D, attribSort_2)

# The attributes input can also be a single primitive attribute sort
attribute_sort_1_3D = attributeSort(lineSegSort3D.name + '_' +  weightSort.name, lineSegSort3D, weightSort)
attribute_sort_2_3D = attributeSort(pointSort3D.name + '_' +  ptDSort.name, pointSort3D, ptDSort)

# Other variants with non-parametric/parametric 2D/3D geometric base sorts can also be created
attribute_sort_1_P3D = attributeSort(lineSegSortP3D.name + '_' +  weightSort.name, lineSegSortP3D, weightSort)
attribute_sort_2_P3D = attributeSort(pointSortP3D.name + '_' +  ptDSort.name, pointSortP3D, ptDSort)

attribute_sort_1_2D = attributeSort(lineSegSort2D.name + '_' +  weightSort.name, lineSegSort2D, weightSort)
attribute_sort_2_2D = attributeSort(pointSort2D.name + '_' +  ptDSort.name, pointSort2D, ptDSort)

attribute_sort_1_P2D = attributeSort(lineSegSortP2D.name + '_' +  weightSort.name, lineSegSortP2D, weightSort)
attribute_sort_2_P2D = attributeSort(pointSortP2D.name + '_' +  ptDSort.name, pointSortP2D, ptDSort)

"""
==========================
CREATING DISJUNCTIVE SORTS
==========================
Disjunctive sorts can be composed of a mixture of primitive and attribute sorts. 
Several types of disjunctive sorts are defined below; naming is not fixed and can be adapted to the user's naming conventions.
However, each disjunctive sort should have a unique name as it will be remembered by the sortal library by its name.
"""
components_3D = [attribute_sort_1_d, attribute_sort_2_d, attribute_sort_1_3D, attribute_sort_2_3D, pointSort3D, planeSegSort3D]
disjName = 'rhino_shapes_3D'
disjSort_3D = disjunctiveSort(disjName, components_3D)      # If already in sort register, retrieves sort

components_P3D = [attribute_sort_1_P3D, attribute_sort_2_P3D]
disjName = 'rhino_shapes_P3D'
disjSort_P3D = disjunctiveSort(disjName, components_P3D)      # If already in sort register, retrieves sort

components_2D = [attribute_sort_1_2D, attribute_sort_2_2D]
disjName = 'rhino_shapes_2D'
disjSort_2D = disjunctiveSort(disjName, components_2D)      # If already in sort register, retrieves sort

components_P2D = [attribute_sort_1_P2D, attribute_sort_2_P2D]
disjName = 'rhino_shapes_P2D'
disjSort_P2D = disjunctiveSort(disjName, components_P2D)      # If already in sort register, retrieves sort

"""
========================
CREATING A COMPOUND SORT
========================
"""
comp_name = 'rhino_shapes_c'
components = [disjSort_P2D, disjSort_2D, disjSort_3D, disjSort_P3D]
compSort = compoundSort(comp_name, components)
print (compSort.type)

