# -*- coding: utf-8 -*-
"""
Created on Tue Mar 28 13:17:49 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .plane import plane2D, plane3D
from ..struct.coordinate import c2, pi
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError

class circle2D(plane2D):
    type = 'ind-circle2D'
    def __init__(self, sort, center, p1, p2 = None, *args, **kwargs):
        super(circle2D, self).__init__(sort)     
        self.center = center
        self.squareRadius = p1.subtract(self.center).squareLength()
        self.radius = self.squareRadius.sqrt()
        
    def isNil(self):
        return False
        
    def getPoints(self):
        s = self.center.add(vector2D(self.radius, 0))
        e = self.center.add(vector2D(0, self.radius))
        return  self.center, s, e
        
    def printVal(self):
        c, s, e = self.getPoints()
        return 'center:'+str(c)+', p1:'+str(s)+', p2:'+str(e)
        
    def printRhino(self, dX = 0, dY = 0):
        return '_Circle '+str(self.center.x + dX)+','+str(self.center.y + dY)+'  '+str(self.radius)
        
    def toString(self, printKey):
        st = ''
        if printKey and self.key is not None:
            st += str(self.key)+' '
        c, s, e = self.getPoints()
        return st + '<'+str(c)+','+str(s)+','+str(e)+'>'
        
    def equals(self, other):
        if self.center == other.center:
            return self.radius == other.radius
        return False
        
    def containsPoint(self, point):
        if point.subtract(self.center).squareLength() == self.squareRadius:
            return True
        return False
        
    def compareValue(self, other):
        c = self.center.compareValue(other.center)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        if other.type != 'ind-circle2D':
            return False
        return self.center == other.center
        
    def diameter(self):
        return self.radius * 2
        
    def circumference (self):
        return self.radius * 2 * pi
        
    def area(self):
        return self.squareRadius * pi
        
    def transformInd(self, transformSet):        
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            newCircle = self.duplicateInd()
        else:
            c, s, e = self.getPoints()
            center = c.transform(affine)
            p1 = s.transform(affine)
            newCircle = circle2D(self.sort, center, p1)
            
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            center = transf.apply(self.center)
            newPts = []
            for i in transf.vects:
                if newCircle.containsPoint(i[0]):
                    newPts.append(i[1])
            newCircle = circle2D(self.sort, center, newPts[0], newPts[-1])
        return newCircle
        
    def replicateInd(self):
        c, s, e = self.getPoints()
        return circle2D(self.sort, c, s)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        if reader.newToken() == ',': # bracket (current token would be ',')
            x,y = vector2D.readVector(reader)
            p2 = vector2D(x,y)
            if reader.token() != '>':
                raise ParsingError(reader,'">" expected')
            reader.newToken() #next token
        else:
            p2 = None
        return circle2D(sort, center, p1, p2)
        
class circle3D(plane3D):
    type = 'ind-circle3D'
    
    def __init__(self, sort, center, p1, p2, *args, **kwargs):
        super(circle3D, self).__init__(sort, [center, p1, p2])   
        self.squareRadius = p1.subtract(center).squareLength()
        self.radius = self.squareRadius.sqrt()
        self.center = center
        self.localCenter = self.localize(center)
        
    def isNil(self):
        return False
        
    def printVal(self):
        c, s, e = self.getPoints()
        return 'center:'+str(c)+', p1:'+str(s)+', p2:'+str(e)
    
    def printRhino(self, dX = 0, dY = 0):
        pCenter = self.center
        pNormal = pCenter.add(self.normal)
        return '_Circle '+str(pCenter.x+dX)+','+str(pCenter.y + dY)+','+str(pCenter.z)+' O '+ str(pNormal.x+dX)+','+str(pNormal.y + dY)+','+str(pNormal.z)+' '+str(self.radius)
        
    def toString(self, printKey):
        st = ''
        if printKey and self.key is not None:
            st += str(self.key)+' '
        c, s, e = self.getPoints()
        return st + '<'+str(c)+','+str(s)+','+str(e)+'>'
        
    def getPoints(self):
        s = self.localCenter.add(vector2D(self.radius, 0))
        e = self.localCenter.add(vector2D(0, self.radius))
        return  self.center, self.position(s), self.position(e)

    def equals(self, other):
        if self.normal == other.normal and self.localCenter == self.localCenter and self.radius == other.radius :
            return True
        return False
        
    def containsPoint(self, point):
        if point == self.center or not point.subtract(self.center).dotproduct(self.normal) == 0:
            return False
        p = self.localize(point)
        if p.subtract(self.localCenter).squareLength() == self.squareRadius:
            return True
        return False
        
    def compareValue(self, other):
        c = super(circle3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.localCenter.compareValue(other.localCenter)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        if other.type != 'ind-circle3D':
            return False
        if not super(circle3D, self).__eq__(other):
            return False
        return self.localCenter == other.localCenter
        
    def diameter(self):
        return self.radius * 2
        
    def circumference (self):
        return self.radius * 2 * pi
        
    def area(self):
        return self.squareRadius * pi
        
    def transformInd(self, transformSet):        
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newCircle = self.duplicateInd()
        else:
            c,s,e = self.getPoints()
            center = c.transform(affine)
            p1 = s.transform(affine)
            p2 = e.transform(affine)    
            newCircle = circle3D(self.sort, center, p1, p2)
            
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            center = transf.apply(self.center)
            newPts = []
            for i in transf.vects:
                if newCircle.containsPoint(i[0]):
                    newPts.append(i[1])
            newCircle = circle3D(self.sort, center, newPts[0], newPts[-1])
        return newCircle
        
    def replicateInd(self):
        c,s,e = self.getPoints()
        return circle3D(self.sort, c, s, e)
        
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        center = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return circle3D(sort, center, p1, p2)